// contextVarDef.h
//
// The code in this file is part of Pyxplot
// <http://www.pyxplot.org.uk>
//
// Copyright (C) 2006-2012 Dominic Ford <coders@pyxplot.org.uk>
//               2008-2012 Ross Church
//
// $Id: contextVarDef.h 1261 2012-07-11 21:38:05Z dcf21 $
//
// Pyxplot is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// You should have received a copy of the GNU General Public License along with
// Pyxplot; if not, write to the Free Software Foundation, Inc., 51 Franklin
// Street, Fifth Floor, Boston, MA  02110-1301, USA

// ----------------------------------------------------------------------------

#ifndef _CONTEXTVARDEF_H
#define _CONTEXTVARDEF_H 1

#include "userspace/context.h"
#include "userspace/pplObj.h"

void ppl_contextVarHierLookup    (ppl_context *c, int srcLineN, int srcId, char *srcFname, char *linetxt, pplObj *stk, int *stkPos, pplObj **out, int base, int offset);
void ppl_contextVarLookup        (ppl_context *c, char *name, pplObj **output, int returnGlobObjs);
void ppl_contextGetVarPointer    (ppl_context *c, char *name, pplObj **output, pplObj *temp);
void ppl_contextRestoreVarPointer(ppl_context *c, char *name, pplObj *temp);

#endif

